/***************************************************************************
mondo-prep-general.c  -  description
-----------------

begin: Fri Apr 19 16:40:35 EDT 2002
copyright : (C) 2002 Mondo  Hugo Rabson
email     : Hugo Rabson <hugorabson@msn.com>
edited by : by Stan Benoit 4/2002
email     : troff@nakedsoul.org
cvsid     : $Id: mondo-prep.c,v 1.14 2004/06/21 20:20:36 hugo Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* mondo-prep-general.c             Hugo Rabson



09/21
- removed g_fprep

08/20
- moved lots of funcs from mondo-prep-general.c to mondo-prep-linux.c
  and mondo-prep-freebsd.c

07/29
- use parted instead of fdisk

07/20
- when starting RAID, don't do it twice :)
- moved Joshua's new get_phys_size_of_drive() code
  from resize_drive... into get_phys_size_of_drive()

06/29
- make sure software RAID devices are formatted IF user says they're to be
- drivelist is struct now, not char[][]

06/26
- drop make_relevant_partition_bootable(); do it yourself in C (mostly)
- offer to reboot if partition table is locked up by the kernel

06/22
- be sure not to resize non-NTFS images when restoring
- set non-NTFS images' partition types properly

06/19
- shut down all LVMs and VGs before prepping

05/07
- usage of parted2fdisk instead of fdisk alone (ia32/ia64 compatibility)
  BCO

03/31
- rewrote partitioning and formatting code to call fdisk once per disk

10/21/2003
- suspend/resume Newt gui before/after calling do_my_funky_lvm_stuff()

10/20
- rewrote format_everything() - what a mess it was.
  It now does things in three phases:-
  - formats software RAID devices (/dev/md0, etc.)
  - formats and configures LVM devices
  - formats regular partitions (/dev/hda1, /dev/sdb2, etc.) 
    and any LVMs recently prepped

10/07
- use strstr(format, "raid") instead of strcmp(format,"raid") to determin
  if partition is a RAID component

09/23
- better comments

09/18
- better logging of RAID activity

05/05
- added Joshua Oreman's FreeBSD patches

04/30
- added textonly mode

04/24
- added lots of assert()'s and log_OS_error()'s

04/21
- format_everything() --- don't let bar go too far
- mkfs -c to check for bad blocks when formatting

04/04
- misc clean-up (Tom Mortell)

01/15/2003
- added code for LVM and SW Raid (Brian Borgeson)

12/10/2002
- line 1238: friendlier output

11/20
- when wiping a given device in preparation for partitioning + formatting
  it, don't wipe the MBR; just the partition table. That allows for
  stupid-ass Compaq users who like to play with their MBR's.
- disable mountlist.txt-->mountlist.txt.pre-resize copying (superfluous)

09/09
- run_program_and_log_output() now takes boolean operator to specify
  whether it will log its activities in the event of _success_

07/01 - 07/31
- added hooks to libmondo
- RAID enhancements (Philippe de Muyter)

01/01 - 03/31
- partition_device() will refuse to partition /dev/mdX now (cos it's a
  virtual drive); however, it will return 0 (i.e. not an error)
- make_list_of_drives() will exclude /dev/md* from list
- mkreiserfs -ff instead of -q (Andy Glass)
- don't resize drive if drive not found (or if its size cannot be det'd)
- when generating list of drives from mountlist, skip the 'p' at the end
  of drive name if /dev/ida/ or /dev/cciss/; just do it (Michael Eisenberg)
- don't specify journal size when formatting ext3
  (used to have -Jsize=10 in the call to mkfs)
- handles files >2GB in size
- format interactively, if Interactive Mode
- removed reference to g_tape_size
- when executing /tmp/i-want-my-lvm, only record the error# if the command
  was _not_ an 'insmod' command
- pass partition size to fdisk in Kilobytes now, not Megabytes
- log fdisk's output to /tmp/mondo-restore.log if it fails
- don't try to format partitions of type 'image'
- don't type to set types of 'image' partitions
- if format code is 1 or 2 chars then assume it is a hex string
- took out all '/ /' comments
- don't extrapolate/add partition from RAID dev to mountlist if it's already
  present in mountlist
- less repetitive logging in the event of vacuum-packing of last part'n
- no extrapolation at all: RAID partitions should be listed in mountlist
  already, thanks to either Mindi v0.5x or the mountlist editor itself
- no longer say, 'and logging to...' when setting a partition's type
- don't run mkfs on RAID partitions (/dev/hd*, /dev/sd*); just set type
- the setting of a partition's type now takes place in a separate subroutine
  from the subroutine that actually creates the partition
- no need to set type if 83: 83 is the default (under fdisk)
- turned on '-Wall'; cleaned up some cruft
- if vacuum-packing partition (i.e. size=0MB --> max) then say, "(maximum)"
  not, "(0 MB)"

11/22/2001
- preliminary code review
- created on Nov 22nd, 2001
*/

/**
 * @file
 * Functions for prepping hard drives: partitioning, formatting, etc.
 */


#include "../common/my-stuff.h"
#include "../common/mondostructures.h"
#include "mondoprep.h"
#include "../common/libmondo.h"
#include "mondo-rstr-tools-EXT.h"
#include <sys/ioctl.h>
#include <linux/hdreg.h>
#include <math.h>


#define PARTED_LOG "/tmp/parted.log"


#define ARCHIVES_PATH MNT_CDROM"/archives"
#define MONDO_WAS_HERE "MONDOWOZEREMONDOWOZEREMONDOWOZEREhahahaMOJOJOJO"

//static char cvsid[] = "$Id: mondo-prep.c,v 1.14 2004/06/21 20:20:36 hugo Exp $";





int g_partition_table_locked_up=0;










void wipe_MBRs_and_reboot_if_necessary(struct mountlist_itself *mountlist)
{
    char *command;
    char *tmp;
    int lino;
    int i;
    FILE*fout;
    char *buf;
    const int blocksize=512;        
        struct list_of_disks *drivelist=NULL;
// If LVMs are present and a zero-and-reboot wasn't recently undertaken
// then zero & insist on reboot.
    malloc_string(command);
    malloc_string(tmp);
    buf = malloc(blocksize);
        if (does_file_exist("/tmp/i-want-my-lvm")) // FIXME - cheating :)
	  {
	    drivelist = malloc(sizeof(struct list_of_disks));
	    make_list_of_drives_in_mountlist(mountlist, drivelist);
	    for (lino = 0; lino < drivelist->entries; lino++)
	      {
	        sprintf(command, "dd if=%s bs=512 count=1 2> /dev/null | grep \"%s\"", drivelist->el[lino].device, MONDO_WAS_HERE);
	        if (!run_program_and_log_output(command, 1))
		  {
		    log_msg(1, "Found MONDO_WAS_HERE marker on drive#%d (%s)", lino, drivelist->el[lino].device);
		    break;
		  }
              }
	    
	    if (lino==drivelist->entries)
	      {
// zero & reboot
	        log_to_screen("I am sorry for the inconvenience but I must ask you to reboot.");
		log_to_screen("I need to reset the Master Boot Record; in order to be");
		log_to_screen("sure the kernel notices, I must reboot after doing it.");
		log_to_screen("Please hit 'Enter' to reboot.");
	        for (lino = 0; lino < drivelist->entries; lino++)
	          {
		    for(i=0;i<blocksize; i++)
		      { buf[i] = 0; }
		    sprintf(buf, "%s\n", MONDO_WAS_HERE);
		    fout = fopen(drivelist->el[lino].device, "w+");
		    if (!fout)
		      {
		        log_msg(1, "Unable to open+wipe %s", drivelist->el[lino].device);
		      }
		    else
		      {
		        if (1!=fwrite(buf, blocksize, 1, fout))
			  {
			    log_msg(1, "Failed to wipe %s", drivelist->el[lino].device);
			  }
			else
			  {
			    log_msg(1, "Successfully wiped %s", drivelist->el[lino].device);
			  }
		        fclose(fout);
	              }
		  }
                system("sync"); system("sync"); system("sync");
		popup_and_OK("I must now reboot. Please leave the boot media in the drive and repeat your actions - e.g. type 'nuke' - and it should work fine.");
                system("reboot"); 
	      }
          }
	  
// Still here? Cool!
  paranoid_free(command);
  paranoid_free(tmp);
  log_msg(1, "Cool. I didn't have to wipe anything.");
}






int fput_string_one_char_at_a_time(FILE*fout, char*str)
{
  int i, j;
  FILE*fq;

  if (ferror(fout))
    {
	return(-1);
    }
  log_msg(5, "Writing string '%s', one char at a time", str);
  j = strlen(str);
  for(i=0; i<j; i++)
    {
      log_msg(6, "Writing %d ('%c')", str[i], str[i]);
      if ((fq = fopen(PARTED_LOG, "a+"))) { fputc(str[i], fq);fclose(fq); }
      fputc(str[i], fout);
      fflush(fout);
      if (str[i]<32) { usleep(1000L*10L); }
    }
  log_msg(5, "Returning");
  sync();
  usleep(1000L*10L);
   
   return(i);
}












/**
 * Format all drives (except those excluded by format_device()) in @p mountlist.
 * @param mountlist The mountlist containing partitions to be formatted.
 * @param interactively If TRUE, then prompt the user before each partition.
 * @return The number of errors encountered (0 for success).
 */
int format_everything(struct mountlist_itself *mountlist, bool interactively)
{
	/** int **************************************************************/
	int retval = 0;
	int lino;
	int res;
//	int i;
//	struct list_of_disks *drivelist;

	/** long *************************************************************/
	long progress_step;

	/** bools ************************************************************/
	bool do_it;

	/** buffers **********************************************************/
	char *tmp;

	/** pointers *********************************************************/
	struct mountlist_line *me;	// mountlist entry
	/** end **************************************************************/

	assert(mountlist!=NULL);
	malloc_string(tmp);
	sprintf(tmp, "format_everything (mountlist, interactively = %s", (interactively) ? "true" : "false");
	log_it(tmp);
	mvaddstr_and_log_it(g_currentY, 0, "Formatting partitions     ");
	open_progress_form("Formatting partitions", "I am now formatting your hard disk partitions.",
			   "This may take up to five minutes.", "", mountlist->entries + 1);

	progress_step = (mountlist->entries > 0) ? g_maximum_progress / mountlist->entries : 1;
// start soft-raids now (because LVM might depend on them)
// ...and for simplicity's sake, let's format them at the same time :)
	log_msg(1, "Stopping all RAID devices");
	stop_all_raid_devices(mountlist);
	system("sync"); system("sync"); system("sync"); 
	sleep(2);
	log_msg(1, "Prepare soft-RAIDs"); // prep and format too
	for (lino = 0; lino < mountlist->entries; lino++) {
		me = &mountlist->el[lino];	// the current mountlist entry
		log_msg(2, "Examining %s", me->device);
		if (!strncmp(me->device, "/dev/md", 7)) {
			if (interactively) {
				// ask user if we should format the current device
				sprintf(tmp, "Shall I format %s (%s) ?", me->device, me->mountpoint);
				do_it = ask_me_yes_or_no(tmp);
			} else {
				do_it = TRUE;
			}
			if (do_it) {
			// NB: format_device() also stops/starts RAID device if necessary
				retval += format_device(me->device, me->format);
			}
			g_current_progress += progress_step;
		}
	}
	system("sync"); system("sync"); system("sync"); 
	sleep(2);
// This last step is probably necessary
//	log_to_screen("Re-starting software RAIDs...");
//	start_all_raid_devices(mountlist);
//	system("sync"); system("sync"); system("sync"); 
//	sleep(5);
// do LVMs now
	log_msg(1, "Creating LVMs");
	if (does_file_exist("/tmp/i-want-my-lvm")) {
		wait_until_software_raids_are_prepped("/proc/mdstat", 10);
		log_to_screen("Configuring LVM");
                if (!g_text_mode) { newtSuspend(); }
/*
		for(i=0; i<3; i++)
		  {
		    res = do_my_funky_lvm_stuff(FALSE, FALSE);
		    if (!res) { break; }
		    sleep(3);
		    res = do_my_funky_lvm_stuff(TRUE, FALSE);
		    sleep(3);
		  }
		if (res) {
			log_msg(1, "Vacuum-packing...");
*/
			res = do_my_funky_lvm_stuff(FALSE, TRUE);
/*
		}
*/
                if (!g_text_mode) { newtResume(); }
		if (!res) { log_to_screen("LVM initialized OK"); }
		else { log_to_screen("Failed to initialize LVM"); }
		// retval += res;
		if (res) {
			retval ++;
		}
	    sleep(3);
	}

// do regulars at last
	sleep(2); // woo!
        log_msg(1, "Formatting regulars");
	for (lino = 0; lino < mountlist->entries; lino++) {
		me = &mountlist->el[lino];	// the current mountlist entry
		if (!strcmp(me->mountpoint, "image")) {
			sprintf(tmp, "Not formatting %s - it's an image", me->device);
			log_it(tmp);
		} else if (!strcmp(me->format, "raid")) {
			sprintf(tmp, "Not formatting %s - it's a raid-let", me->device);
			log_it(tmp);
			continue;
		} else if (!strcmp(me->format, "lvm")) {
			sprintf(tmp, "Not formatting %s - it's an LVM", me->device);
			log_it(tmp);
			continue;
		} else if (!strncmp(me->device, "/dev/md", 7)) {
			sprintf(tmp, "Already formatted %s - it's a soft-RAID dev", me->device);
			log_it(tmp);
			continue;
		} else if (!does_file_exist(me->device)
					&& strncmp(me->device, "/dev/hd", 7)
					&& strncmp(me->device, "/dev/sd", 7)) {
			sprintf(tmp, "Not formatting %s yet - doesn't exist - probably an LVM", me->device);
			log_it(tmp);
			continue;
		} else {
			if (interactively) {
				// ask user if we should format the current device
				sprintf(tmp, "Shall I format %s (%s) ?", me->device, me->mountpoint);
				do_it = ask_me_yes_or_no(tmp);
			} else {
				do_it = TRUE;
			}

			if (do_it)
				retval += format_device(me->device, me->format);
		}

		// update progress bar
		g_current_progress += progress_step;
	}


	// update progress bar to 100% to compensate for
	// rounding errors of the progress_step calculation
	if (lino >= mountlist->entries)
		g_current_progress = g_maximum_progress;

	close_progress_form();

	if (retval) {
		mvaddstr_and_log_it(g_currentY++, 74, "Failed.");
		log_to_screen("Errors occurred during the formatting of your hard drives.");
	} else {
		mvaddstr_and_log_it(g_currentY++, 74, "Done.");
	}

	sprintf(tmp, "format_everything () - %s", (retval) ? "failed!" : "finished successfully");
	log_it(tmp);

	if (g_partition_table_locked_up > 0)
	  {
	    if (retval > 0 && !interactively)
              {
//123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 
                log_to_screen(
  "Partition table locked up %d times. At least one 'mkfs' (format) command", g_partition_table_locked_up);
                log_to_screen(
  "failed. I think these two events are related. Sometimes, parted's ioctl() call");
                log_to_screen(
  "to refresh its copy of the partition table causes the kernel to lock the ");
                log_to_screen(
  "partition table. I believe this has just happened.");
                if (ask_me_yes_or_no("Please choose 'yes' to reboot and try again; or 'no' to ignore this warning and continue."))
                  {
		    system("sync");
		    system("sync");
		    system("sync");
		    system("reboot"); 
		  }
	      }
	    else
	      {
	        log_to_screen(
"Partition table locked up %d time%c. However, disk formatting succeeded.", g_partition_table_locked_up, (g_partition_table_locked_up==1)?'.':'s');
	      }
	  }
        newtSuspend();
	system("clear");
	newtResume();
	paranoid_free(tmp);
	return (retval);
}


/**
 * Create small dummy partitions to fill in the gaps in partition numbering for @p drivename.
 * Each partition created is 32k in size.
 * @param drivename The drive to create the dummy partitions on.
 * @param devno_we_must_allow_for The lowest-numbered real partition; create
 * dummies up to (this - 1).
 * @return The number of errors encountered (0 for success).
 */
int make_dummy_partitions(FILE*pout_to_parted, char *drivename, int devno_we_must_allow_for)
{
	/** int **************************************************************/
	int current_devno;
	int previous_devno;
	int retval = 0;
	int res;

	/** buffers **********************************************************/
	char *tmp;

	/** end **************************************************************/

	malloc_string(tmp);
	assert_string_is_neither_NULL_nor_zerolength(drivename);

	if (devno_we_must_allow_for >= 5) {
		sprintf(tmp, "Making dummy primary %s%d", drivename, 1);
		log_it(tmp);
		g_maximum_progress++;
		res = partition_device(pout_to_parted,drivename, 1, 0, "ext2", 32000);
		retval += res;
		previous_devno = 1;
		current_devno = 5;
	} else {
		previous_devno = 0;
		current_devno = 1;
	}
	for (; current_devno < devno_we_must_allow_for; current_devno++) {
		sprintf(tmp, "Creating dummy partition %s%d", drivename, current_devno);
		log_it(tmp);
		g_maximum_progress++;
		res = partition_device(pout_to_parted,drivename, current_devno, previous_devno, OSSWAP("ext2", "ufs"), 32000);
		retval += res;
		previous_devno = current_devno;
	}
	paranoid_free(tmp);
	return (previous_devno);
}


/**
 * Decide whether @p mountlist contains any RAID devices.
 * @param mountlist The mountlist to examine.
 * @return TRUE if it does, FALSE if it doesn't.
 */
bool mountlist_contains_raid_devices(struct mountlist_itself * mountlist)
{
	/** int *************************************************************/
	int i;
	int matching = 0;

	/** end **************************************************************/

	assert(mountlist != NULL);

	for (i = 0; i < mountlist->entries; i++) {
		if (strstr(mountlist->el[i].device, RAID_DEVICE_STUB)) {
			matching++;
		}
	}
	if (matching) {
		return (TRUE);
	} else {
		return (FALSE);
	}
}










/**
 * Calculate the probable size of @p drive_name by adding up sizes in
 * @p mountlist.
 * @param mountlist The mountlist to use to calculate the size.
 * @param drive_name The drive to calculate the original size of.
 * @return The size of @p drive_name in kilobytes.
 */
long calc_orig_size_of_drive_from_mountlist(struct mountlist_itself
					    *mountlist, char *drive_name)
{
	/** long ************************************************************/
	long original_size_of_drive;

	/** int *************************************************************/
	int partno;

	/** buffers *********************************************************/
	char *tmp;

	/** end *************************************************************/

	malloc_string(tmp);
	assert(mountlist != NULL);
	assert_string_is_neither_NULL_nor_zerolength(drive_name);

	for (original_size_of_drive = 0, partno = 0; partno < mountlist->entries; partno++) {
		if (strncmp(mountlist->el[partno].device, drive_name, strlen(drive_name)) == 0) {
			original_size_of_drive += mountlist->el[partno].size;
		} else {
			sprintf(tmp, "Skipping %s", mountlist->el[partno].device);
//                      log_to_screen(tmp);
		}
	}
	original_size_of_drive = original_size_of_drive / 1024;
	paranoid_free(tmp);
	return (original_size_of_drive);
}


/**
 * Resize a drive's entries in @p mountlist proportionately to fit its new size.
 * There are a few problems with this function:
 * - It won't work if there was any unallocated space on the user's hard drive
 *   when it was backed up.
 * - It won't work if the user's hard drive lies about its size (more common
 *   than you'd think).
 *
 * @param mountlist The mountlist to use for resizing @p drive_name.
 * @param drive_name The drive to resize.
 */
void resize_drive_proportionately_to_suit_new_drives(struct mountlist_itself
						     *mountlist, char *drive_name)
{
	/**buffers **********************************************************/
	char *tmp;

	/** int *************************************************************/
	int partno, lastpart;
		       /** remove driveno, noof_drives stan benoit apr 2002**/

	/** float ***********************************************************/
	float factor;
	float new_size;
//	float newcylinderno;

	/** long *************************************************************/
	long newsizL;
	long current_size_of_drive = 0;
	long original_size_of_drive = 0;
	long final_size;	/* all in Megabytes */
	struct mountlist_reference *drivemntlist;
	bool all_entries_have_zero_KB_size = TRUE;
	/** structures *******************************************************/

	/** end **************************************************************/

	malloc_string(tmp);
	assert(mountlist != NULL);
	assert_string_is_neither_NULL_nor_zerolength(drive_name);

	if (strlen(drive_name) >= strlen(RAID_DEVICE_STUB)) {
		if (strncmp(drive_name, RAID_DEVICE_STUB, strlen(RAID_DEVICE_STUB)) == 0) {
			paranoid_free(tmp);
			return;
		}
	}

	/*
	   sprintf (tmp, "cp -f %s %s.pre-resize", g_mountlist_fname, g_mountlist_fname);
	   run_program_and_log_output (tmp, FALSE);
	 */

	current_size_of_drive = get_phys_size_of_drive(drive_name);
	
	if (current_size_of_drive <= 0) {
		log_it("Not resizing to match %s - can't find drive", drive_name);
		paranoid_free(tmp);
		return;
	}
	sprintf(tmp, "Expanding entries to suit drive %s (%ld MB)", drive_name, current_size_of_drive);
	log_to_screen(tmp);

	drivemntlist = malloc(sizeof (struct mountlist_reference));
	drivemntlist->el = malloc(sizeof(struct mountlist_line *) * MAX_TAPECATALOG_ENTRIES);

	if (!drivemntlist) {
		fatal_error("Cannot malloc temporary mountlist\n");
	}
	
	create_mountlist_for_drive(mountlist, drive_name, drivemntlist);
	for (partno = 0; partno < drivemntlist->entries; partno++) {
		original_size_of_drive += drivemntlist->el[partno]->size;
	}
	original_size_of_drive = original_size_of_drive / 1024;
	if (original_size_of_drive <= 0) {
		sprintf(tmp, "Cannot resize %s's entries. Drive not found.", drive_name);
		log_to_screen(tmp);
		paranoid_free(tmp);
		return;
	}
	factor = (float) (current_size_of_drive) / (float) (original_size_of_drive);
	sprintf(tmp, "Disk %s was %ld MB; is now %ld MB; factor = %f", drive_name, original_size_of_drive, current_size_of_drive,
		factor);
	log_to_screen(tmp);

	lastpart = drivemntlist->entries-1;
	for (partno = 0; partno < drivemntlist->entries; partno++) {
		/* the 'atoi' thing is to make sure we don't try to resize _images_, whose formats will be numeric */
		if ( !atoi(drivemntlist->el[partno]->format) ) {
			new_size = (float) (drivemntlist->el[partno]->size) * factor;
		} else {
			new_size = drivemntlist->el[partno]->size;
		}
		
		if (!strcmp(drivemntlist->el[partno]->mountpoint, "image")) {
			log_msg(1, "Skipping %s (%s) because it's an image",
							drivemntlist->el[partno]->device,
							drivemntlist->el[partno]->mountpoint);
		} else {
			newsizL = (long) new_size;
			sprintf(tmp, "Changing %s from %lld KB to %ld KB", drivemntlist->el[partno]->device, drivemntlist->el[partno]->size,newsizL);
			log_to_screen(tmp);
			drivemntlist->el[partno]->size = newsizL;
		}
		if (drivemntlist->el[partno]->size > 0)
		  { all_entries_have_zero_KB_size = FALSE; }
	}
	final_size = get_phys_size_of_drive(drive_name);
	sprintf(tmp, "final_size = %ld MB", final_size);
	log_msg(3, tmp);
	if (all_entries_have_zero_KB_size)
	  {
	    log_msg(1, "Warning - after resizing, all %s's partitions are zero-size", drive_name);
	  }
	paranoid_free(tmp);
}


/**
 * Resize all partitions in @p mountlist proportionately (each one
 * grows or shrinks by the same percentage) to fit them into the new
 * drives (presumably different from the old ones).
 * @param mountlist The mountlist to resize the drives in.
 */
void resize_mountlist_proportionately_to_suit_new_drives(struct mountlist_itself
							 *mountlist)
{
	/** buffers *********************************************************/
	struct list_of_disks *drivelist;

	/** int *************************************************************/
	int driveno;

	/** end *************************************************************/

	drivelist = malloc(sizeof(struct list_of_disks));
	assert(mountlist != NULL);

	if (g_mountlist_fname[0] == '\0') {
	    log_it ("resize_mountlist_prop...() - warning - mountlist fname is blank");
	    log_it ("That does NOT affect the functioning of this subroutine.");
	    log_it ("--- Hugo, 2002/11/20");
	}
	log_msg(1, "mountlist[0] --> %s (%lld KB)", mountlist->el[0].device, mountlist->el[0].size);
	iamhere("Resizing mountlist");
	make_list_of_drives_in_mountlist(mountlist, drivelist);
	log_msg(1, "drivelist entries = %d", drivelist->entries);
	iamhere("Back from MLoDiM");
	for (driveno = 0; driveno < drivelist->entries; driveno++) {
		resize_drive_proportionately_to_suit_new_drives(mountlist, drivelist->el[driveno].device);
	}
	log_to_screen("Mountlist adjusted to suit current hard drive(s)");
	log_msg(1, "mountlist[0] --> %s (%lld KB)", mountlist->el[0].device, mountlist->el[0].size);
	paranoid_free(drivelist);
}

/**
 * Create a mountlist_reference structure for @p drive_name in @p mountlist.
 * @param mountlist The complete mountlist to get the drive references from.
 * @param drive_name The drive to put in @p drivemntlist.
 * @param drivemntlist The mountlist_reference structure to put the drive's entries in.
 * @note @p drivemntlist and @p drivemntlist->el must be allocated by the caller.
 * @author Ralph Grewe
 */
void create_mountlist_for_drive(struct mountlist_itself *mountlist, char *drive_name, struct mountlist_reference *drivemntlist) {
	int partno;
	char *tmp_drive_name, *c;

	assert(mountlist != NULL);
	assert(drive_name != NULL);
	assert(drivemntlist != NULL);

	log_msg (1, "Creating list of partitions for drive %s",drive_name);

	tmp_drive_name = strdup (drive_name);
	if (!tmp_drive_name) fatal_error ("Out of memory");
	
	/* devfs devices? */
	c = strrchr (tmp_drive_name, '/');
	if (c && strncmp (c, "/disc", 5) == 0) {
	    /* yup its devfs, change the "disc" to "part" so the existing code works */
	    strcpy (c + 1, "part");
	}
	drivemntlist->entries=0;
	for (partno = 0; partno < mountlist->entries; partno++) {
		if (strncmp (mountlist->el[partno].device, tmp_drive_name, strlen(tmp_drive_name)) == 0) {
			drivemntlist->el[drivemntlist->entries] = &mountlist->el[partno];
			drivemntlist->entries++;
		}
	}
	if (tmp_drive_name)
	    free (tmp_drive_name);
}

/* @} - end of prepGroup */
