/***************************************************************************
mondo-prep-linux.c  -  description
-----------------

begin: Fri Apr 19 16:40:35 EDT 2004
copyright : (C) 2004 Mondo  Hugo Rabson
email     : Hugo Rabson <hugorabson@msn.com>
edited by : by Stan Benoit 4/2002
email     : troff@nakedsoul.org
cvsid     : $Id: mondo-prep.c,v 1.14 2004/06/21 20:20:36 hugo Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* mondo-prep-linux.c             Hugo Rabson


09/21
- tried to make vacuum-packing more effective

08/20
- moved lots of funcs from mondo-prep-general.c to mondo-prep-linux.c
  and mondo-prep-freebsd.c

08/20/2004
- first incarnation
*/

/**
 * @file
 * Functions for prepping hard drives: partitioning, formatting, etc.
 */

#ifndef __FreeBSD__

#include "../common/my-stuff.h"
#include "../common/mondostructures.h"
#include "mondoprep.h"
#include "../common/libmondo.h"
#include "mondo-rstr-tools-EXT.h"
#include <sys/ioctl.h>
#include <linux/hdreg.h>
#include <math.h>


#define PARTED_LOG "/tmp/parted.log"

int set_partition_type (FILE*,const char *, int, const char *, long long);





extern int g_partition_table_locked_up;


extern void wipe_MBRs_and_reboot_if_necessary(struct mountlist_itself *mountlist);
extern int format_everything(struct mountlist_itself *mountlist, bool interactively);
extern int fput_string_one_char_at_a_time(FILE*fout, char*str);
extern void create_mountlist_for_drive(struct mountlist_itself *mountlist, char *drive_name, struct mountlist_reference *drivemntlist);
extern void resize_mountlist_proportionately_to_suit_new_drives(struct mountlist_itself
							 *mountlist);



/**
 * @addtogroup prepGroup
 * @{
 */
/**
 * Execute the commands in /tmp/i-want-my-lvm.
 * These should probably be commands to set up LVM.
 * @return The number of errors encountered (0 for success).
 */
 

int do_my_funky_lvm_stuff(bool just_erase_existing_volumes, bool vacuum_pack)
{
	/**  buffers **********************************************/
	char *tmp;
	char *incoming;
	char *command;
	char *lvscan_sz;
	char *lvremove_sz;
	char *pvscan_sz;
	char *vgscan_sz;
	char *vgcreate_sz;
	char *vgchange_sz;
	char *vgremove_sz;
//	char *do_this_last;

	/** char **************************************************/
	char *p;
	char *q;

	/** int ***************************************************/
	int retval = 0;
	int res = 0;
	int i;
	int lvmversion = 1;
	long extents;
	fpos_t orig_pos;
	
	/** pointers **********************************************/
	FILE *fin;

	/** end *****************************************************/

	if (!(fin = fopen("/tmp/i-want-my-lvm", "r"))) {
		log_OS_error("/tmp/i-want-my-lvm");
		return (1);
	}

	malloc_string(tmp);
	malloc_string(incoming);
	malloc_string(lvscan_sz);
	malloc_string(lvremove_sz);
	malloc_string(vgscan_sz);
	malloc_string(pvscan_sz);
	malloc_string(vgcreate_sz);
	malloc_string(vgchange_sz);
	malloc_string(vgremove_sz);
//	malloc_string(do_this_last); // postpone lvcreate call if necessary
	command = malloc(512);

//	do_this_last[0] = '\0';
iamhere("STARTING");
	log_msg(1, "OK, opened i-want-my-lvm. Shutting down LVM volumes...");
	if (find_home_of_exe("lvm")) // found it :) cool
	  {
	    strcpy(lvscan_sz,   "lvm lvscan");
	    strcpy(lvremove_sz, "lvm lvremove");
	    strcpy(vgscan_sz,   "lvm vgscan");
	    strcpy(pvscan_sz,   "lvm pvscan");
	    strcpy(vgcreate_sz, "lvm vgcreate");
	    strcpy(vgchange_sz, "lvm vgchange");
	    strcpy(vgremove_sz, "lvm vgremove");
	  }
	else
	  {
	    strcpy(lvscan_sz,   "lvscan");
	    strcpy(lvremove_sz, "lvremove");
	    strcpy(vgscan_sz,   "vgscan");
	    strcpy(pvscan_sz,   "pvscan");
	    strcpy(vgcreate_sz, "vgcreate");
	    strcpy(vgchange_sz, "vgchange");
	    strcpy(vgremove_sz, "vgremove");
	  }
	sprintf(command, "for i in `%s | cut -d\"'\" -f2 | sort -r` ; do echo \"Shutting down lv $i\" >> " MONDO_LOGFILE "; %s -f $i; done",
					lvscan_sz, lvremove_sz);
	run_program_and_log_output(command, 5);
	sleep(1);
	sprintf(command, "for i in `%s | grep -i lvm | cut -d'\"' -f2` ; do %s -a n $i ; %s $i; echo \"Shutting down vg $i\" >> " MONDO_LOGFILE "; done; %s -a n", vgscan_sz, vgchange_sz, vgremove_sz, vgremove_sz);
	run_program_and_log_output(command, 5);
	if (just_erase_existing_volumes) {
		paranoid_fclose(fin);
		log_msg(1, "Closed i-want-my-lvm. Finished erasing LVMs.");
		retval=0;
		goto end_of_i_want_my_lvm;
	}

	log_msg(1, "OK, rewound i-want-my-lvm. Doing funky stuff...");
	rewind(fin);
	for (fgets(incoming, 512, fin); !feof(fin); fgets(incoming, 512, fin)) {
		fgetpos(fin, &orig_pos);
		if (incoming[0] != '#') {
			continue;
		}
		if (res && strstr(command, "create") && vacuum_pack)
		  {
		    sleep(2);
		    system("sync"); system("sync"); system("sync");
		  }
		if ((p=strstr(incoming, "vgcreate")))
		  {
// include next line(s) if they end in /dev (cos we've got a broken i-want-my-lvm)
		    for (fgets(tmp, 512, fin); !feof(fin); fgets(tmp, 512, fin)) {
			if (tmp[0] == '#') {
			    fsetpos(fin, &orig_pos);
			    break;
			} else {
			    fgetpos(fin, &orig_pos);
			    strcat(incoming, tmp);
			}
		    }
		    for(q=incoming; *q!='\0'; q++)
		      {
		        if (*q<32) { *q=' '; }
		      }
		    strcpy(tmp, p+strlen("vgcreate")+1);
		    for(q=tmp; *q>32; q++);
		    *q = '\0';
		       log_msg(1, "Deleting old entries at /dev/%s", tmp);
//		       sprintf(command, "%s -f %s", vgremove_sz, tmp);
//		       run_program_and_log_output(command, 1);
		       sprintf(command, "rm -Rf /dev/%s", tmp);
		       run_program_and_log_output(command, 1);
		    run_program_and_log_output(vgscan_sz, 1);
		    run_program_and_log_output(pvscan_sz, 1);
		    log_msg(3, "After working around potentially broken i-want-my-lvm, incoming[] is now '%s'", incoming);
		  }
		for (p = incoming + 1; *p == ' '; p++);
		strcpy(command, p);
		for (p = command; *p != '\0'; p++);
		for (; *(p - 1) < 32; p--);
		*p = '\0';
		res = run_program_and_log_output(command, 5);
		if (res>0 && (p=strstr(command, "lvm ")))
		  {
		    *p = *(p+1) = *(p+2) = ' ';
		    res = run_program_and_log_output(command, 5);
		  }
		log_msg(0, "%s --> %d", command, res);
                if (res>0) {res=1;}
		if (res && strstr(command, "lvcreate") && vacuum_pack)
		  {
		    res=0;
		    if (strstr(command, "lvm lvcreate"))
			lvmversion = 2;
		    log_it("%s... so I'll get creative.", tmp);
		    if (lvmversion == 2)
		      {
		        strcpy(tmp, call_program_and_get_last_line_of_output
		    	    ("cat /var/log/mondo-archive.log | tail -n5 | grep Insufficient | tail -n1"));
		      }
		    else
		      {
		        strcpy(tmp, call_program_and_get_last_line_of_output
		    	    ("cat /var/log/mondo-archive.log | tail -n5 | grep lvcreate | tail -n1"));
		      }
		    for(p=tmp; *p!='\0' && !isdigit(*p); p++);
		    extents = atol(p);
		    log_msg(5, "p='%s' --> extents=%ld", p, extents);
		    p = strstr(command, "-L");
		    if (!p)
		      {
		        log_msg(0, "Fiddlesticks. '%s' returned %d", command, res);
		      }
		    else
		      {
			if (lvmversion == 2)
			  {
			    *p++ = '-'; *p++ = 'l'; *p++ = ' ';
			    for(q=p; *q!=' '; q++) { *q = ' '; }
			    sprintf(p, "%ld", extents);
			    i = strlen(p);
			    *(p+i) = ' ';
			  }
			else
			  {
			    p++; p++; p++;
			    for(q=p; *q!=' '; q++) { *(q-1) = ' '; }
			    sprintf(p, "%ld%c", extents, 'm');
			    i = strlen(p);
			    *(p+i) = ' ';
			  }
		        log_msg(5, "Retrying with '%s'", command);
		        res = run_program_and_log_output(command, 5);
			if (res>0) { res=1; }
		        log_msg(0, "%s --> %d", command, res);
			if (!res)
			  {
			    log_msg(5, "YAY! This time, it succeeded.");
			  }
		      }
		  }
		if (strstr(command, "vgcreate"))
		  {
		    log_msg(0, "In case you're interested...");
		    run_program_and_log_output(vgscan_sz, 1);
		    run_program_and_log_output(pvscan_sz, 1);
		  }
                if (res!=0 && !strstr(command, "insmod")) {
			retval ++;
		}
		sprintf(tmp, "echo \"%s\" >> /tmp/out.sh", command);
		system(tmp);
		sleep(1);
	}
	paranoid_fclose(fin);
	log_msg(1, "Closed i-want-my-lvm. Finished doing funky stuff.");
end_of_i_want_my_lvm:
	paranoid_free(tmp);
	paranoid_free(incoming);
	paranoid_free(command);
	paranoid_free(lvscan_sz);
	paranoid_free(lvremove_sz);
	paranoid_free(vgscan_sz);
	paranoid_free(pvscan_sz);
	paranoid_free(vgcreate_sz);
	paranoid_free(vgchange_sz);
	paranoid_free(vgremove_sz);
//	paranoid_free(do_this_last);
	system("sync"); system("sync"); system("sync"); 
	sleep(1);
iamhere("ENDING");
	if (retval>2)
	  {
	    log_msg(1, "%d errors. I'm reporting this.", retval);
	    return(retval);
	  }
	else
	  {
	    log_msg(1, "Not many errors. Returning 0.");
	    return(0);
	  }
}








/**
 * Format @p device as a @p format filesystem.
 * This will use the format command returned by which_format_command_do_i_need().
 * If @p device is an LVM PV, it will not be formatted, and LVM will be started
 * (if not already done). If it's an imagedev, software RAID component, or
 * (under BSD) swap partition, no format will be done.
 * @param device The device to format.
 * @param format The filesystem type to format it as.
 * @return 0 for success, nonzero for failure.
 */
int format_device(char *device, char *format)
{
	/** int **************************************************************/
	int res;
	int retval = 0;

	/** buffers ***********************************************************/
	char *program;
	char *tmp;

	/** end ****************************************************************/

	malloc_string(program);
	malloc_string(tmp);
	assert_string_is_neither_NULL_nor_zerolength(device);
	assert(format != NULL);

	if (strstr(format, "raid")) {	// do not form RAID disks; do it to /dev/md* instead
		sprintf(tmp, "Not formatting %s (it is a RAID disk)", device);
		log_it(tmp);
		paranoid_free(program); paranoid_free(tmp);
		return (0);
	}
	if (strlen(format) <= 2) {
		sprintf(tmp,
			"%s has a really small format type ('%s') - this is probably a hexadecimal string, which would suggest the partition is an image --- I shouldn't format it",
			device, format);
		log_it(tmp);
		paranoid_free(program); paranoid_free(tmp);
		return (0);
	}
	if (is_this_device_mounted(device)) {
		sprintf(tmp, "%s is mounted - cannot format it       ", device);
		log_to_screen(tmp);
		paranoid_free(program); paranoid_free(tmp);
		return (1);
	}
	if (strstr(device, RAID_DEVICE_STUB)) {
		newtSuspend();
		sprintf(tmp, "Initializing RAID device %s", device);
		log_to_screen(tmp);

// Shouldn't be necessary.
		log_to_screen("Stopping %s", device);
		stop_raid_device(device);
		system("sync"); sleep(1);

		log_msg(1, "Making %s", device);
		sprintf(program, "mkraid --really-force %s", device);
		res = run_program_and_log_output(program, 1);
		log_msg(1, "%s returned %d", program, res);
		system("sync"); sleep(3);
		start_raid_device(device);
		system("sync"); sleep(2);
		
//		log_to_screen("Starting %s", device);
//		sprintf(program, "raidstart %s", device);
//		res = run_program_and_log_output(program, 1);
//		log_msg(1, "%s returned %d", program, res);
//		system("sync"); sleep(1);
		system("sync"); sleep(1);
		newtResume();
	}
//#ifndef __FreeBSD__
//#endif

	if (!strcmp(format, "lvm"))
	  {
	    log_msg(1, "Don't format %s - it's part of an lvm volume", device);
	    paranoid_free(program); paranoid_free(tmp);
	    return(0);
	  }
	res = which_format_command_do_i_need(format, program);
	sprintf(tmp, "%s %s", program, device);
	if (strstr(program, "kludge")) {
		strcat(tmp, " /");
	}
	sprintf(program, "sh -c 'echo -en \"y\\ny\\ny\\n\" | %s'", tmp);
	sprintf(tmp, "Formatting %s as %s", device, format);
	update_progress_form(tmp);
	res = run_program_and_log_output(program, FALSE);
	if (res && strstr(program, "kludge")) {
		sprintf(tmp, "Kludge failed; using regular mkfs.%s to format %s", format, device);
		sprintf(program, "mkfs -t %s -F 32 %s", format, device);
		res = run_program_and_log_output(program, FALSE);
	}
	retval += res;
	if (retval) {
		strcat(tmp, "...failed");
	} else {
		strcat(tmp, "...OK");
	}

	log_to_screen(tmp);
	paranoid_free(program); paranoid_free(tmp);
	system("sync"); sleep(1);
	return (retval);
}




/**
 * Partition @p drivename based on @p mountlist.
 * @param mountlist The mountlist to use to guide the partitioning.
 * @param drivename The drive to partition.
 * @return 0 for success, nonzero for failure.
 */
int partition_drive(struct mountlist_itself *mountlist, char *drivename)
{
	/** int *************************************************************/
	int current_devno;
	int previous_devno = 0;
	int lino;
	int retval = 0;
	int i;
	FILE*pout_to_parted=NULL;

	/** long long *******************************************************/
	long long partsize;

	/** buffers *********************************************************/
	char *device_str;
	char *format;
	char *tmp;

	/** end *************************************************************/

	assert(mountlist != NULL);
	assert_string_is_neither_NULL_nor_zerolength(drivename);

	malloc_string(device_str);
	malloc_string(format);
	malloc_string(tmp);

	log_msg(0, "Partitioning drive %s", drivename);

	make_hole_for_file(PARTED_LOG);
        sprintf(tmp, "parted %s --interactive >> %s 2>> %s", drivename, PARTED_LOG, PARTED_LOG);
	pout_to_parted = popen(tmp, "w");
	if (!pout_to_parted) { 
		log_to_screen("Cannot call parted to configure %s", drivename); 
		paranoid_free(device_str);
		paranoid_free(format);
		paranoid_free(tmp);
		return(1); 
	}
	log_msg(1, "Opened out '%s'", tmp);
	if (!strstr(last_line_of_file(PARTED_LOG), "(parted)")) {
		fput_string_one_char_at_a_time(pout_to_parted, "I\n");
	}
//	popup_and_OK("Press ENTER to continue.");

	for (current_devno = 1; current_devno < 99; current_devno++) {
		build_partition_name(device_str, drivename, current_devno);
		lino = find_device_in_mountlist(mountlist, device_str);

		if (lino < 0) {
			// device not found in mountlist
			continue;
		}

		/* OK, we've found partition /dev/hdxN in mountlist; let's prep it */
		/* For FreeBSD, that is      /dev/adXsY */

		log_it("Found partition %s in mountlist", device_str);
		if (!previous_devno) {

			log_it("Wiping %s's partition table", drivename);
			iamhere("New, kernel-friendly partition remover");
			if (!pout_to_parted) { fatal_error("pout_to_parted is NULL"); }
			fput_string_one_char_at_a_time(pout_to_parted, "print\n");
			sync();
			sleep(1);
			fput_string_one_char_at_a_time(pout_to_parted, "mklabel msdos\n");
/*
			for(i=20;i>0;i--)
			  {
			    fprintf(pout_to_parted, "rm %d\n", i);
			    fflush(pout_to_parted);
			  }
*/
			if (current_devno > 1) {
				previous_devno = make_dummy_partitions(pout_to_parted,drivename, current_devno);
			}
		}

		strcpy(format, mountlist->el[lino].format);
		partsize = mountlist->el[lino].size;


		if (current_devno == 5 && previous_devno == 4) {
			log_to_screen("You must leave at least one partition spare as the Extended partition.");
			paranoid_free(device_str);
			paranoid_free(format);
			paranoid_free(tmp);
			return (1);
		}

		retval += partition_device(pout_to_parted, drivename, current_devno, previous_devno, format, partsize);
		previous_devno = current_devno;
	}

	if (pout_to_parted)
	{
// mark relevant partition as bootable
		for(i=0; i<mountlist->entries && strcmp(mountlist->el[i].mountpoint, "/boot"); i++);
		if (i==mountlist->entries)
		  {
		    for(i=0; i<mountlist->entries && strcmp(mountlist->el[i].mountpoint, "/"); i++);
		  }
		if (i < mountlist->entries)
		  {
		    sprintf(tmp, "set %d boot on\n", i);
		    
		    fput_string_one_char_at_a_time(pout_to_parted, tmp);
		  }
// close parted
		fput_string_one_char_at_a_time(pout_to_parted, "print\n");
		fput_string_one_char_at_a_time(pout_to_parted, "quit\n");
		system("sync");
		paranoid_pclose(pout_to_parted); 
		log_msg(0, "------------------- parted.log looks like this ------------------");
		sprintf(tmp, "cat %s >> %s", PARTED_LOG, MONDO_LOGFILE);
		system(tmp);
		log_msg(0, "------------------- end of parted.log... word! ------------------");
		sprintf(tmp, "tail -n6 %s | fgrep \"16: \"", PARTED_LOG);
		if (!run_program_and_log_output(tmp, 5))
		  {
		    g_partition_table_locked_up++;
		    log_to_screen("A flaw in the Linux kernel has locked the partition table.");
		  }
	}
	paranoid_free(device_str);
	paranoid_free(format);
	paranoid_free(tmp);
	return (retval);
}


/**
 * Create partition number @p partno on @p drive with @p parted.
 * @param drive The drive to create the partition on.
//  * @param partno The partition number of the new partition (1-4 are primary, >=5 is logical).
 * @param prev_partno The partition number of the most recently prepped partition.
 * @param format The filesystem type of this partition (used to set the type).
 * @param partsize The size of the partition in @e bytes.
 * @return 0 for success, nonzero for failure.
 */
int partition_device(FILE*pout_to_parted, const char *drive, int partno, int prev_partno, const char *format, long long partsize)
{
	/** int **************************************************************/
	int retval = 0;
	int res = 0;

	/** buffers **********************************************************/
	char *program;
	char *partition_name;
	char *tmp;
	char *logfile;
	char *output;
        char *size_of_drive_in_fp_megs;
	char *end_of_last_partition_in_fp_megs;
	/** pointers **********************************************************/
	char *p;
	char *part_table_fmt;
	FILE *fout;
	/** logs */
	long long real_start, real_end;

	/** end ***************************************************************/


		
	malloc_string(program);
	malloc_string(partition_name);
	malloc_string(tmp);
	malloc_string(logfile);
	malloc_string(output);
	malloc_string(size_of_drive_in_fp_megs);
	malloc_string(end_of_last_partition_in_fp_megs);
	
	assert_string_is_neither_NULL_nor_zerolength(drive);
	assert(format != NULL);

	log_it("partition_device('%s', %d, %d, '%s', %lld) --- starting", drive, partno, prev_partno, format, partsize);

	sprintf(tmp, "parted %s print | grep %s | head -n1 | cut -d':' -f2 | cut -d' ' -f2 | cut -d'-' -f2", drive, drive);
	strcpy(size_of_drive_in_fp_megs, call_program_and_get_last_line_of_output(tmp));
	log_msg(0, "size_of_drive_in_fp_megs = %s", size_of_drive_in_fp_megs);
	
	sprintf(tmp, "parted %s print | grep \"[0-9]+* *[0-9]+*\" | tail -n1 | tr -s ' ' '\t' | cut -f3", drive);
	strcpy(end_of_last_partition_in_fp_megs, call_program_and_get_last_line_of_output(tmp));
	if (atoll(end_of_last_partition_in_fp_megs)<=0) { strcpy(end_of_last_partition_in_fp_megs, "-1"); }
	log_msg(0, "end_of_last_partition_in_fp_megs = %s", end_of_last_partition_in_fp_megs);
	
	if (!strncmp(drive, RAID_DEVICE_STUB, strlen(RAID_DEVICE_STUB))) {
		sprintf(tmp, "Not partitioning %s - it is a virtual drive", drive);
		log_it(tmp);
		goto endoffunc;
	}
	build_partition_name(partition_name, drive, partno);
	if (partsize <= 0) {
		sprintf(tmp, "Partitioning device %s (max size)", partition_name);
	} else {
		sprintf(tmp, "Partitioning device %s (%lld MB)", partition_name, (long long) partsize / 1024);
	}
	update_progress_form(tmp);
	log_it(tmp);

	if (is_this_device_mounted(partition_name)) {
		sprintf(tmp, "%s is mounted, and should not be partitioned", partition_name);
		log_to_screen(tmp);
		retval++;
		goto endoffunc;
	}
	/*  sprintf(tmp,"Partitioning %s  ",partition_name); */
	/*  mvaddstr_and_log_it(g_currentY+1,30,tmp); */
	p = (char *) strrchr(partition_name, '/');
	sprintf(logfile, "/tmp/parted.log.%s", ++p);
	sprintf(program, "parted --interactive %s >> %s 2>> %s", drive, MONDO_LOGFILE, MONDO_LOGFILE);

	output[0] = '\0';
	/* make it a primary/extended/logical */
	if (partno <= 4) {
		sprintf(output + strlen(output), "mkpart primary");
	} else {
		if (partno == 5) {
			part_table_fmt = which_partition_format(drive);
			/* GPT allows more than 4 primary partitions */
			if ((prev_partno >= 4) && (strcmp(part_table_fmt,"MBR") == 0)) {
				log_to_screen("You need to leave at least one partition free, for 'extended/logical'");
				retval++;
				goto endoffunc;
			} else {
				sprintf(output + strlen(output), "mkpart extended %lld %lld\n", atoll(end_of_last_partition_in_fp_megs)+1, atoll(size_of_drive_in_fp_megs));
			}
		}
		strcat(output + strlen(output), "mkpart logical");
	}
	real_start = atoll(end_of_last_partition_in_fp_megs)+1;
	if (partsize>0) {
		real_end = real_start + partsize/1024;
	} else {
		real_end = atoll(size_of_drive_in_fp_megs);
	}
	sprintf(output + strlen(output), " %lld %lld\n", real_start, real_end);
/*
	if (partsize > 0) {
                if (!strcmp(format, "7")) { log_msg(1, "Adding 512K, just in case"); partsize+=512; }
		sprintf(output + strlen(output), "+%lldK", (long long) (partsize));
	strcat(output + strlen(output), "\n");
	}
*/

	if (pout_to_parted)
	{
		log_msg(1, "Doing the new all-in-one fdisk thing");
		log_msg(1, "output = '%s'", output);
		fput_string_one_char_at_a_time(pout_to_parted, output);
//		fput_string_one_char_at_a_time(pout_to_parted, "print\n");
		strcpy(tmp, last_line_of_file(PARTED_LOG));
		if (strstr(tmp, "(parted)"))
		{
			log_msg(1, "Successfully created %s%d", drive, partno);
		}
		else
		{
			log_msg(1, "last line = %s", tmp);
			log_msg(1, "Failed to create %s%d; sending 'Enter'...", drive, partno);
		}
		if (!retval)
		{
			log_msg(1, "Trying to set %s%d's partition type now", drive, partno);
			while(0!=(retval = set_partition_type(pout_to_parted, drive, partno, format, partsize))
			      && partsize > 32768)
			  {
			    partsize -= 16384;
			    log_msg(1, "Failed. Trying again. (partsize is now %lld KB)", partsize);
			    retval = set_partition_type(pout_to_parted, drive, partno, format, partsize);
			}
		  }
		if (retval) { log_msg(1, "...but failed to set type"); }
	}
	else
	{
		fatal_error("This no longer works. Please write to parted all-at-once. :-)");
	}
	g_current_progress++;
	log_it("partition_device() --- leaving");
endoffunc:
	paranoid_free(size_of_drive_in_fp_megs);
	paranoid_free(end_of_last_partition_in_fp_megs);
	paranoid_free(program);
	paranoid_free(partition_name);
	paranoid_free(tmp);
	paranoid_free(logfile);
	paranoid_free(output);
	return (retval);
}


/**
 * Create all partitions listed in @p mountlist.
 * @param mountlist The mountlist to use to guide the partitioning.
 * @return The number of errors encountered (0 for success).
 * @note This sets the partition types but doesn't actually do the formatting.
 * Use format_everything() for that.
 */
int partition_everything(struct mountlist_itself *mountlist)
{
	/** int ************************************************************/
	int lino;
	int retval = 0;
	int i;
	int res;

	/** buffer *********************************************************/
	struct list_of_disks *drivelist;
	/*  struct mountlist_itself new_mtlist, *mountlist; */

	/** end ************************************************************/

	drivelist = malloc(sizeof(struct list_of_disks));
	assert(mountlist != NULL);

	log_it("partition_everything() --- starting");
	mvaddstr_and_log_it(g_currentY, 0, "Partitioning hard drives      ");
	/*  mountlist=orig_mtlist; */
	if (mountlist_contains_raid_devices(mountlist)) {
		/*      mountlist=&new_mtlist; */
		log_msg(0,"Mountlist, including the partitions incorporated in RAID devices:-");
		for (i = 0; i < mountlist->entries; i++) {
			log_it(mountlist->el[i].device);
		}
		log_msg(0, "End of mountlist.");
	}
	log_msg(0, "Stopping all LVMs, just in case");
        if (!g_text_mode) { newtSuspend(); }
	do_my_funky_lvm_stuff(TRUE, FALSE); // just remove old partitions
        if (!g_text_mode) { newtResume(); }
	log_msg(0, "Stopping all software RAID devices, just in case");
	stop_all_raid_devices(mountlist);
	log_msg(0, "Done.");
	
/*	
	if (does_file_exist("/tmp/i-want-my-lvm"))
	  {
	    wipe_MBRs_and_reboot_if_necessary(mountlist); // i.e. if it wasn't done recently
	  }
*/

	open_progress_form("Partitioning devices", "I am now going to partition all your drives.",
			   "This should not take more than five minutes.", "", mountlist->entries);

	log_msg(1, "mountlist[0] --> %s (%lld KB)", mountlist->el[0].device, mountlist->el[0].size);
	log_msg(1, "www - mountlist entries = %d", mountlist->entries);
	make_list_of_drives_in_mountlist(mountlist, drivelist);
	log_msg(1, "www - drivelist entries = %d", drivelist->entries);
	
	/* partition each drive */
	for (lino = 0; lino < drivelist->entries; lino++) {
		res = partition_drive(mountlist, drivelist->el[lino].device);
		retval += res;
	}
	close_progress_form();
	if (retval) {
		mvaddstr_and_log_it(g_currentY++, 74, "Failed.");
		log_to_screen("Errors occurred during the partitioning of your hard drives.");
	} else {
		mvaddstr_and_log_it(g_currentY++, 74, "Done.");
		paranoid_system("rm -f /tmp/fdisk*.log 2> /dev/null");
	}
	newtSuspend();
	system("clear");
	newtResume();
	paranoid_free(drivelist);
	return (retval);
}






/**
 * Set the type of partition number @p partno on @p drive to @p format.
 * @param drive The drive to change the type of a partition on.
 * @param partno The partition number on @p drive to change the type of.
 * @param format The filesystem type this partition will eventually contain.
 * @param partsize The size of this partition, in @e bytes (used for vfat
 * type calculations).
 * @return 0 for success, nonzero for failure.
 */
int set_partition_type(FILE *pout_to_parted, const char *drive, int partno, const char *format, long long partsize)
{
	/** buffers *********************************************************/
  char *partition;
  char *command;
  char *output;
  char *tmp;
  char *partcode;
  char *logfile;

	/** pointers *********************************************************/
	char *p;
	FILE *fout;

	/** int **************************************************************/
	int res = 0;

	/** end **************************************************************/

	assert_string_is_neither_NULL_nor_zerolength(drive);
	assert(format != NULL);

  malloc_string(partition);
  malloc_string(command);
  malloc_string(output);
  malloc_string(tmp);
  malloc_string(partcode);
  malloc_string(logfile);

	build_partition_name(partition, drive, partno);
	p = (char *) strrchr(partition, '/');
	sprintf(logfile, "/tmp/fdisk-set-type.%s.log", ++p);
	if (strcmp(format, "swap") == 0) {
		strcpy(partcode, "82");
	} else if (strcmp(format, "vfat") == 0) {
		if (partsize / 1024 > 8192) {
			strcpy(partcode, "c");
		} else {
			strcpy(partcode, "b");
		}
	} else if (strcmp(format, "ext2") == 0 || strcmp(format, "reiserfs") == 0 || strcmp(format, "ext3") == 0
		   || strcmp(format, "xfs") == 0 || strcmp(format, "jfs") == 0) {
		strcpy(partcode, "83");
	} else if (strcmp(format, "minix") == 0) {
		strcpy(partcode, "81");
	} else if (strcmp(format, "raid") == 0) {
		strcpy(partcode, "fd");
	} else if ((strcmp(format, "ufs") == 0)
		   || (strcmp(format, "ffs") == 0)) {	/* raid autodetect */
		strcpy(partcode, "a5");
	} else if (strcmp(format, "lvm") == 0) {
		strcpy(partcode, "8e");
	} else if (format[0] == '\0') {	/* LVM physical partition */
		partcode[0] = '\0';
	} else if (strlen(format) >= 1 && strlen(format) <= 2) {
		strcpy(partcode, format);
	} else {
		/* probably an image */
		sprintf(tmp, "Unknown format ('%s') - using supplied string anyway", format);
		mvaddstr_and_log_it(g_currentY++, 0, tmp);
		strcpy(partcode, format); // was 83
	}
	sprintf(tmp, "Setting %s's type to %s (%s)", partition, format, partcode);
	log_msg(1, tmp);
	if (partcode[0] != '\0' && strcmp(partcode, "83")) {	/* no need to set type if 83: 83 is default */
		
		if (pout_to_parted) {
			sprintf(output+strlen(output), "set %d type 0x%s\n", partno, partcode);
			if (!strcmp(format, "lvm")) {
				sprintf(output+strlen(output), "set %d lvm on\n", partno);
			} else if (!strcmp(format, "raid")) {
				sprintf(output+strlen(output), "set %d raid on\n", partno);
/*			} else if (!strcmp(format, "swap")) {
				sprintf(output+strlen(output), "set %d linux-swap on\n", partno);
*/			}
			fput_string_one_char_at_a_time(pout_to_parted, output);
		}
		else
		{
			fatal_error("This no longer works. FIXME, please.");
			sprintf(output, "t\n%d\n%s\n", partno, partcode);
			strcat(output, "w\n");
			sprintf(command, "parted2fdisk %s >> %s 2>> %s", drive, MONDO_LOGFILE, MONDO_LOGFILE);
			log_msg(5, "output = '%s'", output);
			log_msg(5, "partno=%d; partcode=%s", partno, partcode);
			log_msg(5, "command = '%s'", command);
			fout = popen(command, "w");
			if (!fout) {
				log_OS_error(command);
				res = 1;
			} else {
				res = 0;
				fprintf(fout, output);
				paranoid_pclose(fout);
			}
		}
	if (res)
	  {
	    log_OS_error(command);
	  }	
	}

  paranoid_free(partition);
  paranoid_free(command);
  paranoid_free(output);
  paranoid_free(tmp);
  paranoid_free(partcode);
  paranoid_free(logfile);

  return (res);
}


int start_raid_device(char *raid_device)
{
	/** int *************************************************************/
	int res;
	int retval = 0;

	/** buffers *********************************************************/
	char *program;

	/** end *************************************************************/

	assert_string_is_neither_NULL_nor_zerolength(raid_device);
	malloc_string(program);
	
	sprintf(program, "raidstart %s", raid_device);
//      sprintf (program, "raidstart " RAID_DEVICE_STUB "*");
	log_msg(1, "program = %s", program);
	res = run_program_and_log_output(program, 1);
	if (res) { log_msg(1, "Warning - failed to start RAID device %s", raid_device); }
	retval += res;
	sleep(1);
	return (retval);
}



/**
 * Stop @p raid_device using @p raidstop.
 * @param raid_device The software RAID device to stop.
 * @return 0 for success, nonzero for failure.
 */
int stop_raid_device(char *raid_device)
{
	/** int *************************************************************/
	int res;
	int retval = 0;

	/** buffers *********************************************************/
	char *program;

	/** end *************************************************************/

	assert_string_is_neither_NULL_nor_zerolength(raid_device);
	malloc_string(program);
	
	sprintf(program, "raidstop %s", raid_device);
//      sprintf (program, "raidstop " RAID_DEVICE_STUB "*");
	log_msg(1, "program = %s", program);
	res = run_program_and_log_output(program, 1);
	if (res) { log_msg(1, "Warning - failed to stop RAID device %s", raid_device); }
	retval += res;
	return (retval);
}


int start_all_raid_devices(struct mountlist_itself *mountlist)
{
	int i;
	int retval=0;
	int res;
	
	for(i=0; i<mountlist->entries; i++)
	  {
	    if (!strncmp(mountlist->el[i].device, RAID_DEVICE_STUB, strlen(RAID_DEVICE_STUB)))
	      {
	        log_msg(1, "Starting %s", mountlist->el[i].device);
		res = start_raid_device(mountlist->el[i].device);
		retval += res;
	      }
	  }
	if (retval) { log_msg(1, "Started all s/w raid devices OK"); }
	else { log_msg(1, "Failed to start some/all s/w raid devices"); }
	return(retval);
}

/**
 * Stop all software RAID devices listed in @p mountlist.
 * @param mountlist The mountlist to stop the RAID devices in.
 * @return The number of errors encountered (0 for success).
 * @bug @p mountlist is not used.
 */
int stop_all_raid_devices(struct mountlist_itself *mountlist)
{
	/** int *************************************************************/
	int retval = 0;
	int res;

	/** char ************************************************************/
	char *incoming;
	char *dev;
	/** pointers ********************************************************/
	char *p;
	FILE *fin;
	int i;

	/** end ****************************************************************/

	malloc_string(dev);
	malloc_string(incoming);
	assert(mountlist != NULL);

	for (i = 0; i < 3; i++) {
		fin = fopen("/proc/mdstat", "r");
		if (!fin) {
			log_OS_error("/proc/mdstat");
			paranoid_free(dev);
			paranoid_free(incoming);
			return (1);
		}
		for (fgets(incoming, MAX_STR_LEN - 1, fin); !feof(fin); fgets(incoming, MAX_STR_LEN - 1, fin)) {
			for (p = incoming; *p != '\0' && (*p != 'm' || *(p + 1) != 'd' || !isdigit(*(p + 2))); p++);
			if (*p != '\0') {
				sprintf(dev, "/dev/%s", p);
				for (p = dev; *p > 32; p++);
				*p = '\0';
				res = stop_raid_device(dev);
			}
		}
	}
	paranoid_fclose(fin);
	if (retval) { log_msg(1, "Warning - unable to stop some RAID devices"); }
	paranoid_free(dev);
	paranoid_free(incoming);
	system("sync"); system("sync"); system("sync"); 
	sleep(1);
	return (retval);
}



/**
 * Decide which command we need to use to format a device of type @p format.
 * @param format The filesystem type we are about to format.
 * @param program Where to put the binary name for this format.
 * @return 0 for success, nonzero for failure.
 */
int which_format_command_do_i_need(char *format, char *program)
{
	/** int *************************************************************/
	int res = 0;

	/** buffers *********************************************************/
	char *tmp;

	/** end ***************************************************************/

	malloc_string(tmp);
	assert_string_is_neither_NULL_nor_zerolength(format);
	assert(program != NULL);

	if (strcmp(format, "swap") == 0) {
		strcpy(program, "mkswap");
	} else if (strcmp(format, "vfat") == 0) {
		strcpy(program, "format-and-kludge-vfat");
	} else if (strcmp(format, "reiserfs") == 0) {
		strcpy(program, "mkreiserfs -ff");
	} else if (strcmp(format, "xfs") == 0) {
		strcpy(program, "mkfs.xfs -f -q");
	} else if (strcmp(format, "jfs") == 0) {
		strcpy(program, "mkfs.jfs");
	} else if (strcmp(format, "ext3") == 0) {
		strcpy(program, "mkfs -t ext2 -F -j -q");
	} else if (strcmp(format, "minix") == 0) {
		strcpy(program, "mkfs.minix");
	} else if (strcmp(format, "ext2") == 0) {
		strcpy(program, "mke2fs -F -q");
	} else {
		sprintf(program, "mkfs -t %s -c", format);	// -c checks for bad blocks
		sprintf(tmp, "Unknown format (%s) - assuming '%s' will do", format, program);
		log_it(tmp);
		res = 0;
	}
	paranoid_free(tmp);
	return (res);
}
#endif
